/* kdump.c - kernel crash tool to dump logs to disk */

/* Copyright (c) 2005 - 2009, TeamF1, Inc. */

/*
modification history
--------------------
01b,01Dec09,sg   made changes for netgear boxes
01q,23oct09,nsr  written
*/

/*
DESCRIPTION
The file implements the functionality needed to capture kernel crash logs and
store them to a persistent storage for retrieval later to perform a post crash 
analysis. 

Registers a printk hook to capture all the messages before a crash and writes 
them to the log file once a panic is notified via the kernel notifier chains. 
Softlocks however are handled by another hook that notifies the kdump whenever 
softlock is detected so that we can dump the logs (teamf1_slock_notify_hook).

teamf1_printk_hook       : receives all the messages from printk
teamf1_slock_notify_hook : notifier hook for softlockups
kdump_panic              : panic notification handler

Following are the configuration paramters (passed as module params) available

  kdump_log_file   : the absolute path of the file name to dump the logs to
  kdump_log_lines  : number of printks that needs can/need to be stored

Note that the logs are rotated if the max lines are exceeded .. So we should
start with the line having "-------- kdump logged" which indicates that this is 
where the trace ended.

WRITE_FLASH protects the changes done for writting the logs to flash. A user 
space damemon(kdumpd) sends an ioctl to this module and waits by acquiring a 
sepaphore: kdumpSem, whenever kernel panic happens we release the semaphore 
which results in the compression of the logs and then writting the conf file 
to the flash
*/

/* includes */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/syscalls.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <asm/uaccess.h>
#include <linux/ioctl.h>
#include <linux/notifier.h>

MODULE_LICENSE("GPL");
/* defines */

#define KDUMP_LOG_LEN_PER_LINE 100
#define KDUMP_LOG_LINES        kdump_log_lines
#define KDUMP_LOG_PATH         kdump_log_file
#define KDUMP_LOG_SIZE         KDUMP_LOG_LINES * (KDUMP_LOG_LEN_PER_LINE+1)
#define DEVICE_NAME            "/dev/kdump"
#define DEVICE_MAJOR            240
#define IOCTL_WAIT_FOR_PANIC    _IO(101,0)
#define WRITE_FLASH
/* typedefs */

/* globals */

char *  kdump_log_file = NULL;
int     kdump_log_lines = 200;

module_param(kdump_log_file, charp, 0600);
module_param(kdump_log_lines, int, 0600);

MODULE_PARM_DESC(kdump_log_file,  "Log gets dumped this this file");
MODULE_PARM_DESC(kdump_log_lines, "Max Lines allowed per log");

char * kdump_logbuf = NULL;
int    bufoffset = 0;
static int const gz_magic[2] = {0x1f, 0x8b}; /* gzip magic header */

/* externs */

extern void (*teamf1_printk_hook) (char * fmt, va_list args);
extern void (*teamf1_panic_printk_hook) (char * fmt, va_list args);

static struct semaphore kdumpSem;
/* forward declarations */

int kdump_panic(struct notifier_block *this, unsigned long event, void *ptr);


static struct notifier_block panic_block = {
    .notifier_call = kdump_panic,
};

int kdumpOpen (struct inode * inode, struct file * file);
int kdumpRelease (struct inode * inode, struct file * file);
int kdumpIoctl (struct inode * inode, struct file * file,
                                         unsigned int cmd, unsigned long arg);
long kdumpCompatIoctl (struct file * file, unsigned int cmd,
                                                 unsigned long arg);
 struct file_operations kdumpops =
    {
    .open = kdumpOpen,
    .release = kdumpRelease,
    .ioctl = kdumpIoctl,
#if defined (CONFIG_COMPAT)
    .compat_ioctl = kdumpCompatIoctl,
#endif /* CONFIG_COMPAT */
    };

int kdumpOpen
    (
     struct inode * inode, 
     struct file *file
    )
    {
    try_module_get(THIS_MODULE); 
    return 0;
    }

int kdumpRelease 
    (
     struct inode * inode, 
     struct file* file
    )
    {
    module_put(THIS_MODULE);
    return 0;
    }

int kdumpIoctl 
    (
     struct inode * inode, 
     struct file* file, 
     unsigned int cmd, 
     unsigned long arg
    )
    {
    switch(cmd)
        {
        case IOCTL_WAIT_FOR_PANIC:
            down_interruptible(&kdumpSem);
            break;
        default:
            printk("Unknown Ioctl\n");
        }
    return 0;
    }

long kdumpCompatIoctl
    (
     struct file *file,
     unsigned int cmd,
     unsigned long arg
    )
    {
    return kdumpIoctl(file->f_dentry->d_inode, file, cmd, arg);
    }


/*******************************************************************************
*
* kdump_panic - receives notification of panic to flush logs to file
*
* This function is registered to receive notifications on panics from the kernel
* and on receiving the event, writes the logs to the file name specified. If the
* file exists, it appends an incremented index to the file name and stores the
* logs in there.
* Also, gets invoked via the teamf1_slock_notify_hook to dump logs during a
* softlockup detection.
* 
* RETURNS: NOTIFY_DONE
*/
int
kdump_panic(struct notifier_block *this, unsigned long event, void *ptr)
{
    struct file *file;
    mm_segment_t old_fs;
    loff_t       pos = 0;
    int          fd;

    if (kdump_log_file == NULL)
        return NOTIFY_DONE;

    old_fs = get_fs();
    set_fs(KERNEL_DS);

    /* open file and write the log to it */
    fd = sys_open(KDUMP_LOG_PATH, O_WRONLY|O_APPEND, 0644);
    file = fget(fd);

    if (file) 
        {
        vfs_write(file, kdump_logbuf, KDUMP_LOG_SIZE, &pos);
        fput(file);
        }
    sys_close(fd);

#ifdef WRITE_FLASH
    /*release the semaphore*/
    up(&kdumpSem);
#endif
    set_fs(old_fs);

    return NOTIFY_DONE;
}

/*******************************************************************************
*
* kdump_add_printk - adds the printk messages to kdump log buffer
*
* This is hook registered to receive all messages from printk kernel function
* and store them in a circular log with the limit defined the module param
* kdump_log_lines. If the logs are too many, beyond the max lines allowed the 
* log file is rotated and reused.
* 
* RETURNS: NONE
*/

void kdump_add_printk (char * fmt, va_list args)
    {
    char * tempbuf = kdump_logbuf + bufoffset;

    memset (tempbuf, 0x0, (KDUMP_LOG_LEN_PER_LINE+1));


    *tempbuf = '#';
    /* add message to buffer */
    vsnprintf (tempbuf + 1, KDUMP_LOG_LEN_PER_LINE, fmt, args);

    bufoffset += KDUMP_LOG_LEN_PER_LINE +1;

    /* rotate logs if we are out of buffers */
    if ((KDUMP_LOG_SIZE - bufoffset) < KDUMP_LOG_LEN_PER_LINE)
        bufoffset = 0;
    }

/*******************************************************************************
*
* kdump_init - initialize the log buffer and registers the hooks and notifier
*
* Creates a log buffer based on the log size specified, registers the kernel
* hooks and the panic notifiction handler with the kernel notifier chains.
*
* RETURNS: NONE
*/

static int __init kdump_init(void)
    {
    int retVal=-1;

    kdump_logbuf = kmalloc (KDUMP_LOG_SIZE, GFP_ATOMIC);

    if (kdump_logbuf == NULL)
        return 0;
#ifdef WRITE_FLASH
    /* initialize semaphore */
    sema_init (&kdumpSem, 0);

    /* register char device /dev/kdump */
    retVal = register_chrdev (DEVICE_MAJOR, DEVICE_NAME, &kdumpops);
    if (retVal < 0)
        {
        printk (KERN_ERR "registering char device failed");
        return (-1);
        }
#endif

    memset (kdump_logbuf, 0x0, KDUMP_LOG_SIZE);

    teamf1_panic_printk_hook = kdump_add_printk;

    notifier_chain_register(&panic_notifier_list, &panic_block);

    return 0;
    }

/*******************************************************************************
*
* kdump_exit - deinitialize the log buffer and unregisters the hooks, notifier
*
* Frees the allocated kdump log buffer and unhooks the printk and soflockup
* callers and unregisters itself from the panic notification chain.
*
* RETURNS: NONE
*/

static void __exit kdump_exit(void)
    {
    kfree (kdump_logbuf);

    teamf1_printk_hook = NULL;

#ifdef WRITE_FLASH
    /* unregister char device */
     unregister_chrdev (DEVICE_MAJOR, DEVICE_NAME);
#endif

    notifier_chain_unregister(&panic_notifier_list, &panic_block);
    }

module_init(kdump_init);
module_exit(kdump_exit);


